@@ -1,5 +1,5 @@
 /* GSequencer - Advanced GTK Sequencer
- * Copyright (C) 2005-2018 Joël Krähemann
+ * Copyright (C) 2005-2022 Joël Krähemann
  *
  * This file is part of GSequencer.
  *
@@ -25,10 +25,11 @@
 #include <ags/audio/ags_buffer.h>
 #include <ags/audio/ags_audio_signal.h>
 #include <ags/audio/ags_audio_buffer_util.h>
+#include <ags/audio/ags_resample_util.h>
 
 #include <math.h>
 
-void ags_sound_resource_base_init(AgsSoundResourceInterface *interface);
+void ags_sound_resource_base_init(AgsSoundResourceInterface *ginterface);
 
 /**
  * SECTION:ags_sound_resource
@@ -44,9 +45,11 @@ void ags_sound_resource_base_init(AgsSoundResourceInterface *interface);
 GType
 ags_sound_resource_get_type()
 {
-  static GType ags_type_sound_resource = 0;
+  static volatile gsize g_define_type_id__volatile = 0;
+
+  if(g_once_init_enter (&g_define_type_id__volatile)){
+    GType ags_type_sound_resource = 0;
 
-  if(!ags_type_sound_resource){
     static const GTypeInfo ags_sound_resource_info = {
       sizeof(AgsSoundResourceInterface),
       (GBaseInitFunc) ags_sound_resource_base_init,
@@ -56,14 +59,16 @@ ags_sound_resource_get_type()
     ags_type_sound_resource = g_type_register_static(G_TYPE_INTERFACE,
 						     "AgsSoundResource", &ags_sound_resource_info,
 						     0);
+
+    g_once_init_leave(&g_define_type_id__volatile, ags_type_sound_resource);
   }
 
-  return(ags_type_sound_resource);
+  return g_define_type_id__volatile;
 }
 
 
 void
-ags_sound_resource_base_init(AgsSoundResourceInterface *interface)
+ags_sound_resource_base_init(AgsSoundResourceInterface *ginterface)
 {
   /* empty */
 }
@@ -77,7 +82,7 @@ ags_sound_resource_base_init(AgsSoundResourceInterface *interface)
  * 
  * Returns: %TRUE if operation was successful, otherwise %FALSE.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 gboolean
 ags_sound_resource_open(AgsSoundResource *sound_resource,
@@ -101,14 +106,16 @@ ags_sound_resource_open(AgsSoundResource *sound_resource,
  * ags_sound_resource_rw_open:
  * @sound_resource: the #AgsSoundResource
  * @filename: the filename as string
- * @create: create
+ * @audio_channels: the audio channels count
+ * @samplerate: the samplerate
+ * @create: if %TRUE create file, otherwise don't
  * 
  * Open @sound_resource for reading/writing and assign filename. Setting
  * @create to %TRUE causes to create the file if it doesn't exist.
  * 
  * Returns: %TRUE if operation was successful, otherwise %FALSE.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 gboolean
 ags_sound_resource_rw_open(AgsSoundResource *sound_resource,
@@ -138,7 +145,7 @@ ags_sound_resource_rw_open(AgsSoundResource *sound_resource,
  * 
  * Load audio data of @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_load(AgsSoundResource *sound_resource)
@@ -160,7 +167,7 @@ ags_sound_resource_load(AgsSoundResource *sound_resource)
  * 
  * Get information about @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_info(AgsSoundResource *sound_resource,
@@ -187,7 +194,7 @@ ags_sound_resource_info(AgsSoundResource *sound_resource,
  * 
  * Set presets of @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_set_presets(AgsSoundResource *sound_resource,
@@ -218,7 +225,7 @@ ags_sound_resource_set_presets(AgsSoundResource *sound_resource,
  * 
  * Get presets of @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_get_presets(AgsSoundResource *sound_resource,
@@ -253,7 +260,7 @@ ags_sound_resource_get_presets(AgsSoundResource *sound_resource,
  * 
  * Returns: the count of frames actually read
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 guint
 ags_sound_resource_read(AgsSoundResource *sound_resource,
@@ -289,7 +296,7 @@ ags_sound_resource_read(AgsSoundResource *sound_resource,
  * Write @sbuffer to @sound_resource @frame_count number of 
  * frames having @format by skipping @saudio_channels.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_write(AgsSoundResource *sound_resource,
@@ -314,7 +321,7 @@ ags_sound_resource_write(AgsSoundResource *sound_resource,
  * 
  * Flush @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_flush(AgsSoundResource *sound_resource)
@@ -335,7 +342,7 @@ ags_sound_resource_flush(AgsSoundResource *sound_resource)
  * 
  * Seek the @sound_resource @frame_count from @whence.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_seek(AgsSoundResource *sound_resource,
@@ -356,7 +363,7 @@ ags_sound_resource_seek(AgsSoundResource *sound_resource,
  * 
  * Close @sound_resource.
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_sound_resource_close(AgsSoundResource *sound_resource)
@@ -377,9 +384,9 @@ ags_sound_resource_close(AgsSoundResource *sound_resource)
  * 
  * Read audio signal from @sound_resource.
  * 
- * Returns: a #GList-struct containing #AgsAudioSignal
+ * Returns: (element-type AgsAudio.AudioSignal) (transfer full): a #GList-struct containing #AgsAudioSignal
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 GList*
 ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
@@ -411,7 +418,7 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
 				 &audio_channels,
 				 &samplerate,
 				 &buffer_size,
-				 NULL);
+				 &format);
   
   g_object_get(G_OBJECT(sound_resource),
 	       "format", &format,
@@ -442,15 +449,26 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
     i_start = audio_channel;
     i_stop = i_start + 1;
   }
+  
+  copy_mode = ags_audio_buffer_util_get_copy_mode(ags_audio_buffer_util_format_from_soundcard(target_format),
+						  ags_audio_buffer_util_format_from_soundcard(format));
 
   data = NULL;
-
+  target_data = NULL;
+  
   if(samplerate != target_samplerate){
+    buffer_size = (guint) ceil((double) target_buffer_size / (double) target_samplerate * (double) samplerate);
+    
+    ags_sound_resource_set_presets(AGS_SOUND_RESOURCE(sound_resource),
+				   audio_channels,
+				   samplerate,
+				   buffer_size,
+				   format);
+    
     data = ags_stream_alloc(buffer_size,
 			    format);
-    
-    copy_mode = ags_audio_buffer_util_get_copy_mode(ags_audio_buffer_util_format_from_soundcard(target_format),
-						    ags_audio_buffer_util_format_from_soundcard(format));
+    target_data = ags_stream_alloc(target_buffer_size,
+				   format);
   }
     
   for(i = i_start; i < i_stop; i++){
@@ -465,28 +483,39 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
 					NULL,
 					NULL);
     g_object_set(audio_signal,
-		 "samplerate", samplerate, //TODO:JK: implement me target_samplerate
+		 "frame-count", frame_count,
+		 "samplerate", target_samplerate,
 		 "buffer-size", target_buffer_size,
 		 "format", target_format,
-		 "loop-start", loop_start,
-		 "loop-end", loop_end,
+		 "loop-start", target_samplerate * (loop_start / samplerate),
+		 "loop-end", target_samplerate * (loop_end / samplerate),
 		 NULL);
     ags_audio_signal_stream_resize(audio_signal,
 				   (guint) ceil(frame_count / target_buffer_size) + 1);
+    audio_signal->length = (guint) ceil(frame_count / target_buffer_size) + 1;
     audio_signal->stream_current = audio_signal->stream;
-
+    
     start_list = g_list_prepend(start_list,
 				audio_signal);
 
     stream = audio_signal->stream;
 
-    if(samplerate != target_samplerate){
-      //TODO:JK: implement me
-      g_warning("samplerate != target_samplerate");
-    }
+    g_object_set(audio_signal,
+		 "last-frame", frame_count,
+		 NULL);
     
     while(stream != NULL){
       if(samplerate != target_samplerate){
+	AgsResampleUtil resample_util;
+
+	guint allocated_buffer_length;
+
+	allocated_buffer_length = buffer_size;
+
+	if(allocated_buffer_length < target_buffer_size){
+	  allocated_buffer_length = target_buffer_size;
+	}
+
 	if(format == AGS_SOUNDCARD_DOUBLE){
 	  ags_audio_buffer_util_clear_double(data, 1,
 					     buffer_size);
@@ -503,16 +532,39 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
 				i,
 				buffer_size, format);
 
-	target_data = ags_audio_buffer_util_resample(data, 1,
-						     format, samplerate,
-						     buffer_size,
-						     target_samplerate);
+	ags_audio_buffer_util_clear_buffer(target_data, 1,
+					   target_buffer_size, ags_audio_buffer_util_format_from_soundcard(format));
+
+	resample_util.secret_rabbit.src_ratio = samplerate / target_samplerate;
+
+	resample_util.secret_rabbit.input_frames = buffer_size;
+	resample_util.secret_rabbit.data_in = g_malloc(allocated_buffer_length * sizeof(gfloat));
+
+	resample_util.secret_rabbit.output_frames = target_buffer_size;
+	resample_util.secret_rabbit.data_out = g_malloc(allocated_buffer_length * sizeof(gfloat));
+  
+	resample_util.destination = target_data;
+	resample_util.destination_stride = 1;
+
+	resample_util.source = data;
+	resample_util.source_stride = 1;
+
+	resample_util.buffer_length = allocated_buffer_length;
+	resample_util.format = format;
+	resample_util.samplerate = samplerate;
+
+	resample_util.audio_buffer_util_format = ags_audio_buffer_util_format_from_soundcard(format);
+  
+	resample_util.target_samplerate = target_samplerate;
+
+	ags_resample_util_compute(&resample_util);  
+
+	g_free(resample_util.secret_rabbit.data_out);
+	g_free(resample_util.secret_rabbit.data_in);
 
 	ags_audio_buffer_util_copy_buffer_to_buffer(stream->data, 1, 0,
 						    target_data, 1, 0,
 						    target_buffer_size, copy_mode);
-
-	free(target_data);
       }else{
 	ags_sound_resource_read(AGS_SOUND_RESOURCE(sound_resource),
 				stream->data, 1,
@@ -529,7 +581,15 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
     free(data);
   }
 
+  if(target_data != NULL){
+    free(target_data);
+  }
+  
   start_list = g_list_reverse(start_list);
+
+  g_list_foreach(start_list,
+		 (GFunc) g_object_ref,
+		 NULL);
   
   return(start_list);
 }
@@ -545,9 +605,9 @@ ags_sound_resource_read_audio_signal(AgsSoundResource *sound_resource,
  * 
  * Read wave from @sound_resource.
  * 
- * Returns: a #GList-struct containing #AgsWave
+ * Returns: (element-type AgsAudio.Wave) (transfer full): a #GList-struct containing #AgsWave
  * 
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 GList*
 ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
@@ -561,6 +621,9 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
   void *target_data, *data;
 
   guint copy_mode;
+  guint64 relative_offset;
+  guint64 x_point_offset;
+  guint64 current_offset;
   guint frame_count;
   guint audio_channels;
   guint target_samplerate, samplerate;
@@ -575,12 +638,12 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
   ags_sound_resource_info(AGS_SOUND_RESOURCE(sound_resource),
 			  &frame_count,
 			  NULL, NULL);
-
+  
   ags_sound_resource_get_presets(AGS_SOUND_RESOURCE(sound_resource),
 				 &audio_channels,
 				 &samplerate,
 				 &buffer_size,
-				 NULL);
+				 &format);
 
   g_object_get(G_OBJECT(sound_resource),
 	       "format", &format,
@@ -607,31 +670,35 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
   if(audio_channel == -1){
     i_start = 0;
     i_stop = audio_channels;
-
-    //    g_message("do %d", audio_channels);
   }else{
     i_start = audio_channel;
     i_stop = i_start + 1;
   }
 
+  copy_mode = ags_audio_buffer_util_get_copy_mode(ags_audio_buffer_util_format_from_soundcard(target_format),
+						  ags_audio_buffer_util_format_from_soundcard(format));
+  
   data = NULL;
-
+  target_data = NULL;
+  
   if(samplerate != target_samplerate){
+    buffer_size = (guint) ceil((double) target_buffer_size / (double) target_samplerate * (double) samplerate);
+    
+    ags_sound_resource_set_presets(AGS_SOUND_RESOURCE(sound_resource),
+				   audio_channels,
+				   samplerate,
+				   buffer_size,
+				   format);
+    
     data = ags_stream_alloc(buffer_size,
 			    format);
-    
-    copy_mode = ags_audio_buffer_util_get_copy_mode(ags_audio_buffer_util_format_from_soundcard(target_format),
-						    ags_audio_buffer_util_format_from_soundcard(format));
+    target_data = ags_stream_alloc(target_buffer_size,
+				   format);
   }
   
   for(i = i_start; i < i_stop; i++){
     AgsWave *wave;
 
-    AgsTimestamp *timestamp;
-
-    guint64 offset;
-    guint64 relative_offset;
-    guint frame_count;
     gboolean success;
 
     ags_sound_resource_seek(AGS_SOUND_RESOURCE(sound_resource),
@@ -647,42 +714,51 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
 
     start_list = ags_wave_add(start_list,
 			      wave);
-
-
-    g_object_get(wave,
-		 "timestamp", &timestamp,
-		 NULL);
     
-    relative_offset = AGS_WAVE_DEFAULT_BUFFER_LENGTH * samplerate;
+    relative_offset = AGS_WAVE_DEFAULT_BUFFER_LENGTH * target_samplerate;
     
-    offset = x_offset;
-    frame_count = target_buffer_size;
-    
-    success = TRUE;
-    
-    while(success){
-      AgsBuffer *buffer;
+    x_point_offset = x_offset;
       
+    current_offset = 0;
+    
+    while(current_offset < frame_count){
+      AgsBuffer *buffer;
+
+      guint read_count;
       guint num_read;
       gboolean create_wave;
       
       create_wave = FALSE;
 
-      if(offset + frame_count >= ags_timestamp_get_ags_offset(timestamp) + relative_offset){
-	create_wave = TRUE;
+      read_count = target_buffer_size;
+  
+      if(x_point_offset + read_count > relative_offset * floor(x_point_offset / relative_offset) + relative_offset){
+	read_count = relative_offset * floor((x_point_offset + read_count) / relative_offset) - x_point_offset;
 
-	frame_count = ags_timestamp_get_ags_offset(timestamp) + relative_offset - offset;
+	create_wave = TRUE;
+      }else if(x_point_offset + read_count == relative_offset * floor(x_point_offset / relative_offset) + relative_offset){
+	create_wave = TRUE;
       }
       
       buffer = ags_buffer_new();
       g_object_set(buffer,
-		   "x", offset,
+		   "x", x_point_offset,
 		   "samplerate", target_samplerate,
 		   "buffer-size", target_buffer_size,
 		   "format", target_format,
 		   NULL);
 
       if(samplerate != target_samplerate){
+	AgsResampleUtil resample_util;
+
+	guint allocated_buffer_length;
+
+	allocated_buffer_length = buffer_size;
+
+	if(allocated_buffer_length < target_buffer_size){
+	  allocated_buffer_length = target_buffer_size;
+	}
+
 	if(format == AGS_SOUNDCARD_DOUBLE){
 	  ags_audio_buffer_util_clear_double(data, 1,
 					     buffer_size);
@@ -699,29 +775,58 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
 					   i,
 					   buffer_size, format);
 
-	target_data = ags_audio_buffer_util_resample(data, 1,
-						     format, samplerate,
-						     buffer_size,
-						     target_samplerate);
+	ags_audio_buffer_util_clear_buffer(target_data, 1,
+					   target_buffer_size, ags_audio_buffer_util_format_from_soundcard(format));
+	
+	resample_util.secret_rabbit.src_ratio = samplerate / target_samplerate;
+
+	resample_util.secret_rabbit.input_frames = buffer_size;
+	resample_util.secret_rabbit.data_in = g_malloc(allocated_buffer_length * sizeof(gfloat));
+
+	resample_util.secret_rabbit.output_frames = target_buffer_size;
+	resample_util.secret_rabbit.data_out = g_malloc(allocated_buffer_length * sizeof(gfloat));
+  
+	resample_util.destination = target_data;
+	resample_util.destination_stride = 1;
+
+	resample_util.source = data;
+	resample_util.source_stride = 1;
+
+	resample_util.buffer_length = allocated_buffer_length;
+	resample_util.format = format;
+	resample_util.samplerate = samplerate;
+
+	resample_util.audio_buffer_util_format = ags_audio_buffer_util_format_from_soundcard(format);
+  
+	resample_util.target_samplerate = target_samplerate;
+
+	ags_resample_util_compute(&resample_util);  
+
+	g_free(resample_util.secret_rabbit.data_out);
+	g_free(resample_util.secret_rabbit.data_in);
 
 	ags_audio_buffer_util_copy_buffer_to_buffer(buffer->data, 1, 0,
 						    target_data, 1, 0,
-						    frame_count, copy_mode);
-
-	free(target_data);
+						    target_buffer_size, copy_mode);
       }else{
 	num_read = ags_sound_resource_read(AGS_SOUND_RESOURCE(sound_resource),
 					   buffer->data, 1,
 					   i,
-					   frame_count, target_format);
+					   read_count, target_format);
       }
-      //      g_message("read %d[%d-%d]: %d", frame_count, i, i_stop, num_read);
+      //      g_message("read %d[%d-%d]: %d", read_count, i, i_stop, num_read);
 
+      if(num_read == 0){
+	break;
+      }
+      
       ags_wave_add_buffer(wave,
 			  buffer,
 			  FALSE);
       
       if(create_wave){
+	AgsTimestamp *timestamp;
+
 	wave = ags_wave_new(NULL,
 			    i);
 	g_object_set(wave,
@@ -730,30 +835,37 @@ ags_sound_resource_read_wave(AgsSoundResource *sound_resource,
 		     "format", target_format,
 		     NULL);
 
-	start_list = ags_wave_add(start_list,
-				  wave);
-
 	g_object_get(wave,
 		     "timestamp", &timestamp,
 		     NULL);
 	ags_timestamp_set_ags_offset(timestamp,
-				     offset);
+				     (guint64) relative_offset * floor((x_point_offset + read_count) / relative_offset));
+	
+	g_object_unref(timestamp);
+
+	start_list = ags_wave_add(start_list,
+				  wave);
 
-	frame_count = target_buffer_size;
       }
-      
+            
       /* iterate */
-      offset += target_buffer_size;
+      x_point_offset += read_count;
 
-      if(num_read != buffer_size){
-	success = FALSE;
-      }
+      current_offset += num_read;
     }
   }
-
+  
   if(data != NULL){
     free(data);
   }
   
+  if(target_data != NULL){
+    free(target_data);
+  }  
+
+  g_list_foreach(start_list,
+		 (GFunc) g_object_ref,
+		 NULL);
+  
   return(start_list);
 }
