@@ -1,5 +1,5 @@
 /* GSequencer - Advanced GTK Sequencer
- * Copyright (C) 2005-2018 Joël Krähemann
+ * Copyright (C) 2005-2021 Joël Krähemann
  *
  * This file is part of GSequencer.
  *
@@ -21,7 +21,8 @@
 
 #include <ags/lib/ags_log.h>
 
-#include <ags/object/ags_marshal.h>
+#include <ags/file/ags_file.h>
+
 #include <ags/object/ags_connectable.h>
 
 #include <gio/gio.h>
@@ -30,7 +31,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <pwd.h>
 
 #include <ags/i18n.h>
 
@@ -94,7 +94,7 @@ enum{
 enum{
   PROP_0,
   PROP_MAIN_LOOP,
-  PROP_TASK_THREAD,
+  PROP_TASK_LAUNCHER,
   PROP_CONFIG,
   PROP_FILE,
 };
@@ -104,14 +104,14 @@ static guint application_context_signals[LAST_SIGNAL];
 
 AgsApplicationContext *ags_application_context = NULL;
 
-static pthread_mutex_t ags_application_context_class_mutex = PTHREAD_MUTEX_INITIALIZER;
-
 GType
 ags_application_context_get_type()
 {
-  static GType ags_type_application_context = 0;
+  static volatile gsize g_define_type_id__volatile = 0;
+
+  if(g_once_init_enter (&g_define_type_id__volatile)){
+    GType ags_type_application_context = 0;
 
-  if(!ags_type_application_context){
     static const GTypeInfo ags_application_context_info = {
       sizeof (AgsApplicationContextClass),
       NULL, /* base_init */
@@ -138,15 +138,39 @@ ags_application_context_get_type()
     g_type_add_interface_static(ags_type_application_context,
 				AGS_TYPE_CONNECTABLE,
 				&ags_connectable_interface_info);
+
+    g_once_init_leave(&g_define_type_id__volatile, ags_type_application_context);
   }
 
-  return (ags_type_application_context);
+  return g_define_type_id__volatile;
+}
+
+GType
+ags_application_context_flags_get_type()
+{
+  static volatile gsize g_flags_type_id__volatile;
+
+  if(g_once_init_enter (&g_flags_type_id__volatile)){
+    static const GFlagsValue values[] = {
+      { AGS_APPLICATION_CONTEXT_ADDED_TO_REGISTRY, "AGS_APPLICATION_CONTEXT_ADDED_TO_REGISTRY", "application-context-added-to-registry" },
+      { AGS_APPLICATION_CONTEXT_CONNECTED, "AGS_APPLICATION_CONTEXT_CONNECTED", "application-context-connected" },
+      { AGS_APPLICATION_CONTEXT_TYPES_REGISTERED, "AGS_APPLICATION_CONTEXT_TYPES_REGISTERED", "application-context-types-registered" },
+      { 0, NULL, NULL }
+    };
+
+    GType g_flags_type_id = g_flags_register_static(g_intern_static_string("AgsApplicationContextFlags"), values);
+
+    g_once_init_leave (&g_flags_type_id__volatile, g_flags_type_id);
+  }
+  
+  return g_flags_type_id__volatile;
 }
 
 void
 ags_application_context_class_init(AgsApplicationContextClass *application_context)
 {
   GObjectClass *gobject;
+
   GParamSpec *param_spec;
 
   ags_application_context_parent_class = g_type_class_peek_parent(application_context);
@@ -166,7 +190,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    *
    * The assigned main-loop.
    * 
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   param_spec = g_param_spec_object("main-loop",
 				   i18n_pspec("main loop of application context"),
@@ -178,19 +202,19 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
 				  param_spec);
 
   /**
-   * AgsApplicationContext:task-thread:
+   * AgsApplicationContext:task-launcher:
    *
-   * The assigned task thread.
+   * The assigned task launcher.
    * 
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
-  param_spec = g_param_spec_object("task-thread",
-				   i18n_pspec("task thread"),
-				   i18n_pspec("The task thread"),
+  param_spec = g_param_spec_object("task-launcher",
+				   i18n_pspec("task launcher"),
+				   i18n_pspec("The task launcher"),
 				   G_TYPE_OBJECT,
 				   G_PARAM_READABLE | G_PARAM_WRITABLE);
   g_object_class_install_property(gobject,
-				  PROP_TASK_THREAD,
+				  PROP_TASK_LAUNCHER,
 				  param_spec);
 
   /**
@@ -198,7 +222,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    *
    * The assigned config.
    * 
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   param_spec = g_param_spec_object("config",
 				   i18n_pspec("config of application context"),
@@ -214,7 +238,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    *
    * The assigned file.
    * 
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   param_spec = g_param_spec_object("file",
 				   i18n_pspec("file of application context"),
@@ -233,6 +257,9 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
 
   application_context->register_types = ags_application_context_real_register_types;
   
+  application_context->read = NULL;
+  application_context->write = NULL;
+
   application_context->quit = ags_application_context_real_quit;
 
   /* signals */
@@ -242,7 +269,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    *
    * The ::load-config notifies to load configuration.
    *
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   application_context_signals[LOAD_CONFIG] =
     g_signal_new("load-config",
@@ -260,7 +287,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    * The ::prepare signal should be implemented to prepare
    * your application context.
    *
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   application_context_signals[PREPARE] =
     g_signal_new("prepare",
@@ -278,7 +305,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    * The ::setup signal should be implemented to setup
    * your application context.
    *
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   application_context_signals[SETUP] =
     g_signal_new("setup",
@@ -296,7 +323,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    * The ::register-types signal should be implemented to load
    * your types.
    *
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   application_context_signals[REGISTER_TYPES] =
     g_signal_new("register-types",
@@ -313,7 +340,7 @@ ags_application_context_class_init(AgsApplicationContextClass *application_conte
    *
    * The ::quit notifies to load configuration.
    *
-   * Since: 2.0.0
+   * Since: 3.0.0
    */
   application_context_signals[QUIT] =
     g_signal_new("quit",
@@ -351,24 +378,10 @@ void
 ags_application_context_init(AgsApplicationContext *application_context)
 {
   GFile *file;
-  struct passwd *pw;
 
   application_context->flags = 0;
-
-  application_context->obj_mutexattr = (pthread_mutexattr_t *) malloc(sizeof(pthread_mutexattr_t));
-
-  pthread_mutexattr_init(application_context->obj_mutexattr);
-  pthread_mutexattr_settype(application_context->obj_mutexattr,
-			    PTHREAD_MUTEX_RECURSIVE);
-
-#ifdef __linux__
-  pthread_mutexattr_setprotocol(application_context->obj_mutexattr,
-				PTHREAD_PRIO_INHERIT);
-#endif
-
   
-  application_context->obj_mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
-  pthread_mutex_init(application_context->obj_mutex, application_context->obj_mutexattr);
+  g_rec_mutex_init(&(application_context->obj_mutex));
 
   /* uuid */
   application_context->uuid = ags_uuid_alloc();
@@ -377,19 +390,22 @@ ags_application_context_init(AgsApplicationContext *application_context)
   application_context->argc = 0;
   application_context->argv = NULL;
   
-  application_context->version = AGS_VERSION;
-  application_context->build_id = AGS_BUILD_ID;
+  application_context->version = g_strdup(AGS_VERSION);
+  application_context->build_id = g_strdup(AGS_BUILD_ID);
 
-  application_context->log = NULL;  
+  g_atomic_int_set(&(application_context->is_ready),
+		   FALSE);
+  
+  application_context->log = NULL;
+  
   application_context->domain = NULL;
+  
   application_context->config = NULL;
 
   application_context->main_loop = NULL;
-  application_context->task_thread = NULL;
-  application_context->autosave_thread = NULL;
-  
-  application_context->file = NULL;
+  application_context->task_launcher = NULL;
   
+  application_context->file = NULL;  
   application_context->history = NULL;
 }
 
@@ -401,16 +417,12 @@ ags_application_context_set_property(GObject *gobject,
 {
   AgsApplicationContext *application_context;
 
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
   
   application_context = AGS_APPLICATION_CONTEXT(gobject);
 
   /* get application context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
   
   switch(prop_id){
   case PROP_MAIN_LOOP:
@@ -419,10 +431,10 @@ ags_application_context_set_property(GObject *gobject,
       
       main_loop = (GObject *) g_value_get_object(value);
 
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
       
       if(main_loop == application_context->main_loop){  
-	pthread_mutex_unlock(application_context_mutex);
+	g_rec_mutex_unlock(application_context_mutex);
 
 	return;
       }
@@ -437,34 +449,34 @@ ags_application_context_set_property(GObject *gobject,
       
       application_context->main_loop = main_loop;
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
-  case PROP_TASK_THREAD:
+  case PROP_TASK_LAUNCHER:
     {
-      GObject *task_thread;
+      GObject *task_launcher;
       
-      task_thread = (GObject *) g_value_get_object(value);
+      task_launcher = (GObject *) g_value_get_object(value);
 
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
       
-      if(task_thread == application_context->task_thread){  
-	pthread_mutex_unlock(application_context_mutex);
+      if(task_launcher == application_context->task_launcher){  
+	g_rec_mutex_unlock(application_context_mutex);
 
 	return;
       }
 
-      if(application_context->task_thread != NULL){
-	g_object_unref(application_context->task_thread);
+      if(application_context->task_launcher != NULL){
+	g_object_unref(application_context->task_launcher);
       }
       
-      if(task_thread != NULL){
-	g_object_ref(G_OBJECT(task_thread));
+      if(task_launcher != NULL){
+	g_object_ref(G_OBJECT(task_launcher));
       }
       
-      application_context->task_thread = task_thread;
+      application_context->task_launcher = task_launcher;
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   case PROP_CONFIG:
@@ -473,10 +485,10 @@ ags_application_context_set_property(GObject *gobject,
       
       config = (AgsConfig *) g_value_get_object(value);
 
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
       
       if(config == application_context->config){  
-	pthread_mutex_unlock(application_context_mutex);
+	g_rec_mutex_unlock(application_context_mutex);
 
 	return;
       }
@@ -491,7 +503,7 @@ ags_application_context_set_property(GObject *gobject,
       
       application_context->config = config;
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   case PROP_FILE:
@@ -500,10 +512,10 @@ ags_application_context_set_property(GObject *gobject,
       
       file = (AgsFile *) g_value_get_object(value);
 
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
 
       if(file == application_context->file){
-	pthread_mutex_unlock(application_context_mutex);
+	g_rec_mutex_unlock(application_context_mutex);
 	
 	return;
       }
@@ -518,7 +530,7 @@ ags_application_context_set_property(GObject *gobject,
       
       application_context->file = (AgsFile *) file;
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   default:
@@ -536,52 +548,48 @@ ags_application_context_get_property(GObject *gobject,
 {
   AgsApplicationContext *application_context;
 
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
   
   application_context = AGS_APPLICATION_CONTEXT(gobject);
 
   /* get application context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   switch(prop_id){
   case PROP_MAIN_LOOP:
     {
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
 
       g_value_set_object(value, application_context->main_loop);
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
-  case PROP_TASK_THREAD:
+  case PROP_TASK_LAUNCHER:
     {
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
 
-      g_value_set_object(value, application_context->task_thread);
+      g_value_set_object(value, application_context->task_launcher);
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   case PROP_CONFIG:
     {
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
 
       g_value_set_object(value, application_context->config);
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   case PROP_FILE:
     {
-      pthread_mutex_lock(application_context_mutex);
+      g_rec_mutex_lock(application_context_mutex);
 
       g_value_set_object(value, application_context->file);
   
-      pthread_mutex_unlock(application_context_mutex);
+      g_rec_mutex_unlock(application_context_mutex);
     }
     break;
   default:
@@ -605,11 +613,7 @@ ags_application_context_dispose(GObject *gobject)
   }
 
   /* config */
-  if(application_context->config != NULL){
-    g_object_set(application_context->config,
-		 "application-context", NULL,
-		 NULL);
-    
+  if(application_context->config != NULL){    
     g_object_unref(application_context->config);
 
     application_context->config = NULL;
@@ -617,39 +621,20 @@ ags_application_context_dispose(GObject *gobject)
   
   /* main loop */
   if(application_context->main_loop != NULL){
-    g_object_set(application_context->main_loop,
-		 "application-context", NULL,
-		 NULL);
-
     g_object_unref(application_context->main_loop);
 
     application_context->main_loop = NULL;
   }
 
-  /* autosave thread */
-  if(application_context->autosave_thread != NULL){
-    g_object_set(application_context->autosave_thread,
-		 "application-context", NULL,
-		 NULL);
+  /* task launcher */
+  if(application_context->task_launcher != NULL){
+    g_object_unref(application_context->task_launcher);
 
-    g_object_unref(application_context->autosave_thread);
-
-    application_context->autosave_thread = NULL;
-  }
-
-  /* task thread */
-  if(application_context->task_thread != NULL){
-    g_object_unref(application_context->task_thread);
-
-    application_context->task_thread = NULL;
+    application_context->task_launcher = NULL;
   }
 
   /* file */
   if(application_context->file != NULL){
-    g_object_set(application_context->file,
-		 "application-context", NULL,
-		 NULL);
-
     g_object_unref(application_context->file);
 
     application_context->file = NULL;
@@ -666,12 +651,8 @@ ags_application_context_finalize(GObject *gobject)
 
   application_context = AGS_APPLICATION_CONTEXT(gobject);
 
-  /* application mutex */
-  pthread_mutexattr_destroy(application_context->obj_mutexattr);
-  free(application_context->obj_mutexattr);
-
-  pthread_mutex_destroy(application_context->obj_mutex);
-  free(application_context->obj_mutex);
+  g_free(application_context->version);
+  g_free(application_context->build_id);
 
   /* log */
   if(application_context->log != NULL){
@@ -680,34 +661,17 @@ ags_application_context_finalize(GObject *gobject)
 
   /* config */
   if(application_context->config != NULL){
-    g_object_set(application_context->config,
-		 "application-context", NULL,
-		 NULL);
-
     g_object_unref(application_context->config);
   }
   
   /* main loop */
   if(application_context->main_loop != NULL){
-    g_object_set(application_context->main_loop,
-		 "application-context", NULL,
-		 NULL);
-
     g_object_unref(application_context->main_loop);
   }
 
-  /* autosave thread */
-  if(application_context->autosave_thread != NULL){
-    g_object_set(application_context->autosave_thread,
-		 "application-context", NULL,
-		 NULL);
-
-    g_object_unref(application_context->autosave_thread);
-  }
-
-  /* task thread */
-  if(application_context->task_thread != NULL){
-    g_object_unref(application_context->task_thread);
+  /* task launcher */
+  if(application_context->task_launcher != NULL){
+    g_object_unref(application_context->task_launcher);
   }
 
   /* file */
@@ -730,23 +694,19 @@ ags_application_context_get_uuid(AgsConnectable *connectable)
   
   AgsUUID *ptr;
 
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   application_context = AGS_APPLICATION_CONTEXT(connectable);
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* get UUID */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   ptr = application_context->uuid;
 
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
   
   return(ptr);
 }
@@ -764,23 +724,19 @@ ags_application_context_is_ready(AgsConnectable *connectable)
   
   gboolean is_ready;
 
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   application_context = AGS_APPLICATION_CONTEXT(connectable);
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* check is added */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   is_ready = (((AGS_APPLICATION_CONTEXT_ADDED_TO_REGISTRY & (application_context->flags)) != 0) ? TRUE: FALSE);
 
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
   
   return(is_ready);
 }
@@ -851,23 +807,19 @@ ags_application_context_is_connected(AgsConnectable *connectable)
   
   gboolean is_connected;
 
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   application_context = AGS_APPLICATION_CONTEXT(connectable);
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* check is connected */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   is_connected = (((AGS_APPLICATION_CONTEXT_CONNECTED & (application_context->flags)) != 0) ? TRUE: FALSE);
   
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
   
   return(is_connected);
 }
@@ -900,21 +852,6 @@ ags_application_context_disconnect(AgsConnectable *connectable)
   ags_application_context_unset_flags(application_context, AGS_APPLICATION_CONTEXT_CONNECTED);
 }
 
-/**
- * ags_application_context_get_class_mutex:
- * 
- * Use this function's returned mutex to access mutex fields.
- *
- * Returns: the class mutex
- * 
- * Since: 2.0.0
- */
-pthread_mutex_t*
-ags_application_context_get_class_mutex()
-{
-  return(&ags_application_context_class_mutex);
-}
-
 /**
  * ags_application_context_test_flags:
  * @application_context: the #AgsApplicationContext
@@ -924,32 +861,28 @@ ags_application_context_get_class_mutex()
  * 
  * Returns: %TRUE if flags are set, else %FALSE
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 gboolean
 ags_application_context_test_flags(AgsApplicationContext *application_context, guint flags)
 {
   gboolean retval;  
   
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   if(!AGS_IS_APPLICATION_CONTEXT(application_context)){
     return(FALSE);
   }
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* test */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   retval = (flags & (application_context->flags)) ? TRUE: FALSE;
   
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
 
   return(retval);
 }
@@ -961,32 +894,28 @@ ags_application_context_test_flags(AgsApplicationContext *application_context, g
  *
  * Enable a feature of #AgsApplicationContext.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_set_flags(AgsApplicationContext *application_context, guint flags)
 {
   guint application_context_flags;
   
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   if(!AGS_IS_APPLICATION_CONTEXT(application_context)){
     return;
   }
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* set flags */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   application_context->flags |= flags;
   
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
 }
     
 /**
@@ -996,32 +925,28 @@ ags_application_context_set_flags(AgsApplicationContext *application_context, gu
  *
  * Disable a feature of AgsApplicationContext.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_unset_flags(AgsApplicationContext *application_context, guint flags)
 {
   guint application_context_flags;
   
-  pthread_mutex_t *application_context_mutex;
+  GRecMutex *application_context_mutex;
 
   if(!AGS_IS_APPLICATION_CONTEXT(application_context)){
     return;
   }
 
   /* get application_context mutex */
-  pthread_mutex_lock(ags_application_context_get_class_mutex());
-  
-  application_context_mutex = application_context->obj_mutex;
-  
-  pthread_mutex_unlock(ags_application_context_get_class_mutex());
+  application_context_mutex = AGS_APPLICATION_CONTEXT_GET_OBJ_MUTEX(application_context);
 
   /* unset flags */
-  pthread_mutex_lock(application_context_mutex);
+  g_rec_mutex_lock(application_context_mutex);
 
   application_context->flags &= (~flags);
   
-  pthread_mutex_unlock(application_context_mutex);
+  g_rec_mutex_unlock(application_context_mutex);
 }
 
 void
@@ -1036,7 +961,7 @@ ags_application_context_real_load_config(AgsApplicationContext *application_cont
  *
  * Signal to load and parse configuration.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_load_config(AgsApplicationContext *application_context)
@@ -1047,7 +972,6 @@ ags_application_context_load_config(AgsApplicationContext *application_context)
   g_signal_emit(G_OBJECT(application_context),
 		application_context_signals[LOAD_CONFIG], 0);
   g_object_unref(G_OBJECT(application_context));
-  g_object_unref(application_context);
 }
 
 void
@@ -1062,7 +986,7 @@ ags_application_context_real_prepare(AgsApplicationContext *application_context)
  *
  * Prepare @application_context.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_prepare(AgsApplicationContext *application_context)
@@ -1073,7 +997,6 @@ ags_application_context_prepare(AgsApplicationContext *application_context)
   g_signal_emit(G_OBJECT(application_context),
 		application_context_signals[PREPARE], 0);
   g_object_unref(G_OBJECT(application_context));
-  g_object_unref(application_context);
 }
 
 void
@@ -1088,7 +1011,7 @@ ags_application_context_real_setup(AgsApplicationContext *application_context)
  *
  * Setup @application_context.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_setup(AgsApplicationContext *application_context)
@@ -1099,7 +1022,6 @@ ags_application_context_setup(AgsApplicationContext *application_context)
   g_signal_emit(G_OBJECT(application_context),
 		application_context_signals[SETUP], 0);
   g_object_unref(G_OBJECT(application_context));
-  g_object_unref(application_context);
 }
 
 void
@@ -1114,7 +1036,7 @@ ags_application_context_real_register_types(AgsApplicationContext *application_c
  *
  * Notification to register your types.
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_register_types(AgsApplicationContext *application_context)
@@ -1125,7 +1047,6 @@ ags_application_context_register_types(AgsApplicationContext *application_contex
   g_signal_emit(G_OBJECT(application_context),
 		application_context_signals[REGISTER_TYPES], 0);
   g_object_unref(G_OBJECT(application_context));
-  g_object_unref(application_context);
 }
 
 void
@@ -1141,7 +1062,7 @@ ags_application_context_real_quit(AgsApplicationContext *application_context)
  *
  * Calls exit()
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 void
 ags_application_context_quit(AgsApplicationContext *application_context)
@@ -1159,37 +1080,37 @@ ags_application_context_quit(AgsApplicationContext *application_context)
  * 
  * Get your application context instance.
  *
- * Returns: the #AgsApplicationContext instance
+ * Returns: (transfer none): the #AgsApplicationContext instance
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 AgsApplicationContext*
 ags_application_context_get_instance()
 {
-  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+  static GMutex mutex;
 
-  pthread_mutex_lock(&mutex);
+  g_mutex_lock(&mutex);
   
   if(ags_application_context == NULL){
     ags_application_context = ags_application_context_new(NULL,
 							  NULL);
   }
 
-  pthread_mutex_unlock(&mutex);
+  g_mutex_unlock(&mutex);
   
   return(ags_application_context);
 }
 
 /**
  * ags_application_context_new:
- * @main_loop: the #AgsMainLoop
- * @config: the #AgsConfig
+ * @main_loop: (nullable): the #AgsMainLoop
+ * @config: (nullable): the #AgsConfig
  *
  * Create a new instance of #AgsApplicationContext
  * 
  * Returns: the #AgsApplicationContext instance
  *
- * Since: 2.0.0
+ * Since: 3.0.0
  */
 AgsApplicationContext*
 ags_application_context_new(GObject *main_loop,
@@ -1204,5 +1125,3 @@ ags_application_context_new(GObject *main_loop,
 
   return(application_context);
 }
-
-
